/*
 * Copyright (C) 2025 The pgmoneta community
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors may
 * be used to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <walfile/rm.h>
#include <walfile/rm_gist.h>
#include <utils.h>

struct gist_xlog_delete*
create_gist_xlog_delete(void)
{
   struct gist_xlog_delete* wrapper = malloc(sizeof(struct gist_xlog_delete));

   if (server_config->version >= 16)
   {
      wrapper->parse = pgmoneta_wal_parse_gist_xlog_delete_v16;
      wrapper->format = pgmoneta_wal_format_gist_xlog_delete_v16;
   }
   else
   {
      wrapper->parse = pgmoneta_wal_parse_gist_xlog_delete_v15;
      wrapper->format = pgmoneta_wal_format_gist_xlog_delete_v15;
   }

   return wrapper;
}

// Parsing function for version 15
void
pgmoneta_wal_parse_gist_xlog_delete_v15(struct gist_xlog_delete* wrapper, const void* rec)
{
   memcpy(&wrapper->data.v15, rec, sizeof(struct gist_xlog_delete_v15));
}

// Parsing function for version 16
void
pgmoneta_wal_parse_gist_xlog_delete_v16(struct gist_xlog_delete* wrapper, const void* rec)
{
   memcpy(&wrapper->data.v16, rec, sizeof(struct gist_xlog_delete_v16));
}

// Formatting function for version 15
char*
pgmoneta_wal_format_gist_xlog_delete_v15(struct gist_xlog_delete* wrapper, char* buf)
{
   struct gist_xlog_delete_v15* xlrec = &wrapper->data.v15;
   buf = pgmoneta_format_and_append(buf, "latestRemovedXid: %u; ntodelete: %u", xlrec->latestRemovedXid, xlrec->ntodelete);
   return buf;
}

// Formatting function for version 16
char*
pgmoneta_wal_format_gist_xlog_delete_v16(struct gist_xlog_delete* wrapper, char* buf)
{
   struct gist_xlog_delete_v16* xlrec = &wrapper->data.v16;
   buf = pgmoneta_format_and_append(buf, "delete: snapshot_conflict_horizon_id %u, nitems: %u",
                                    xlrec->snapshotConflictHorizon, xlrec->ntodelete);
   return buf;
}

struct gist_xlog_page_reuse*
create_gist_xlog_page_reuse(void)
{
   struct gist_xlog_page_reuse* wrapper = malloc(sizeof(struct gist_xlog_page_reuse));

   if (server_config->version >= 16)
   {
      wrapper->parse = pgmoneta_wal_parse_gist_xlog_page_reuse_v16;
      wrapper->format = pgmoneta_wal_format_gist_xlog_page_reuse_v16;
   }
   else
   {
      wrapper->parse = pgmoneta_wal_parse_gist_xlog_page_reuse_v15;
      wrapper->format = pgmoneta_wal_format_gist_xlog_page_reuse_v15;
   }

   return wrapper;
}

void
pgmoneta_wal_parse_gist_xlog_page_reuse_v15(struct gist_xlog_page_reuse* wrapper, const void* rec)
{
   memcpy(&wrapper->data.v15, rec, sizeof(struct gist_xlog_page_reuse_v15));
}

void
pgmoneta_wal_parse_gist_xlog_page_reuse_v16(struct gist_xlog_page_reuse* wrapper, const void* rec)
{
   memcpy(&wrapper->data.v16, rec, sizeof(struct gist_xlog_page_reuse_v16));
}

char*
pgmoneta_wal_format_gist_xlog_page_reuse_v15(struct gist_xlog_page_reuse* wrapper, char* buf)
{
   struct gist_xlog_page_reuse_v15* xlrec = &wrapper->data.v15;
   buf = pgmoneta_format_and_append(buf, "rel %u/%u/%u; blk %u; latestRemovedXid %u:%u",
                                    xlrec->node.spcNode, xlrec->node.dbNode,
                                    xlrec->node.relNode, xlrec->block,
                                    EPOCH_FROM_FULL_TRANSACTION_ID(xlrec->latestRemovedFullXid),
                                    XID_FROM_FULL_TRANSACTION_ID(xlrec->latestRemovedFullXid));
   return buf;
}

char*
pgmoneta_wal_format_gist_xlog_page_reuse_v16(struct gist_xlog_page_reuse* wrapper, char* buf)
{
   struct gist_xlog_page_reuse_v16* xlrec = &wrapper->data.v16;
   buf = pgmoneta_format_and_append(buf, "rel %u/%u/%u; blk %u; snapshot_conflict_horizon_id %u:%u",
                                    xlrec->locator.spcOid, xlrec->locator.dbOid,
                                    xlrec->locator.relNumber, xlrec->block,
                                    EPOCH_FROM_FULL_TRANSACTION_ID(xlrec->snapshot_conflict_horizon),
                                    XID_FROM_FULL_TRANSACTION_ID(xlrec->snapshot_conflict_horizon));
   return buf;
}

static char*
out_gistxlogPageUpdate(char* buf, struct gist_xlog_page_update* xlrec)
{
   return buf;
}

static char*
out_gistxlogPageSplit(char* buf, struct gist_xlog_page_split* xlrec)
{
   buf = pgmoneta_format_and_append(buf, "page_split: splits to %d pages",
                                    xlrec->npage);
   return buf;
}

static char*
out_gistxlogPageDelete(char* buf, struct gist_xlog_page_delete* xlrec)
{
   buf = pgmoneta_format_and_append(buf, "deleteXid %u:%u; downlink %u",
                                    EPOCH_FROM_FULL_TRANSACTION_ID(xlrec->deleteXid),
                                    XID_FROM_FULL_TRANSACTION_ID(xlrec->deleteXid),
                                    xlrec->downlinkOffset);
   return buf;
}

char*
pgmoneta_wal_gist_desc(char* buf, struct decoded_xlog_record* record)
{
   char* rec = XLOG_REC_GET_DATA(record);
   uint8_t info = XLOG_REC_GET_INFO(record) & ~XLR_INFO_MASK;
   struct gist_xlog_page_reuse* xlrec_reuse;
   struct gist_xlog_delete* xlrec_delete;

   switch (info)
   {
      case XLOG_GIST_PAGE_UPDATE:
         buf = out_gistxlogPageUpdate(buf, (struct gist_xlog_page_update*) rec);
         break;
      case XLOG_GIST_PAGE_REUSE:
         xlrec_reuse = create_gist_xlog_page_reuse();
         xlrec_reuse->parse(xlrec_reuse, rec);
         buf = xlrec_reuse->format(xlrec_reuse, buf);
         free(xlrec_reuse);
         break;
      case XLOG_GIST_DELETE:
         xlrec_delete = create_gist_xlog_delete();
         xlrec_delete->parse(xlrec_delete, rec);
         buf = xlrec_delete->format(xlrec_delete, buf);
         free(xlrec_delete);
         break;
      case XLOG_GIST_PAGE_SPLIT:
         buf = out_gistxlogPageSplit(buf, (struct gist_xlog_page_split*) rec);
         break;
      case XLOG_GIST_PAGE_DELETE:
         buf = out_gistxlogPageDelete(buf, (struct gist_xlog_page_delete*) rec);
         break;
      case XLOG_GIST_ASSIGN_LSN:
         /* No details to write out */
         break;
   }
   return buf;
}
